/* size+1 for addmul_2 ? */
/* addmul_3 not right */



/* mpn_mullowhalf_basecase -- base case low half multiply.

   Copyright 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"
#include "mpm.h"
#include "mpm-impl.h"


void
mpn_mullowhalf_basecase (mp_ptr wp, mp_srcptr xp, mp_srcptr yp, mp_size_t size)
{
  ASSERT (size >= 1);
  ASSERT (! MPN_OVERLAP_P (wp, 2*size, xp, size));
  ASSERT (! MPN_OVERLAP_P (wp, 2*size, yp, size));

  if (size == 1)
    {
      wp[0] = (xp[0] * yp[0]) & GMP_NUMB_MASK;
      return;
    }

  if (size == 2)
    {
      mp_limb_t  w1, w0;
      mp_limb_t  x0 = xp[0];
      mp_limb_t  y0 = yp[0];
      umul_ppmm (w1, w0, x0, y0 << GMP_NAIL_BITS);
      wp[0] = w0 >> GMP_NAIL_BITS;
      wp[1] = (w1 + x0*yp[1] + xp[1]*y0) & GMP_NUMB_MASK;
      return;
    }

#if HAVE_NATIVE_mpn_mul_2
  mpn_mul_2 (wp, xp, size, yp);
  wp += 2, yp += 2, size -= 2;
#else
  mpn_mul_1 (wp, xp, size, yp[0]);
  wp += 1, yp += 1, size -= 1;
#endif

#if HAVE_NATIVE_mpn_addmul_3
  while (size > 3)
    {
      mpn_addmul_3 (wp, xp, size, yp);
      wp += 3, xp += 3, size -= 3;
    }
  if (LIKELY (size >= 1))
    {
      mp_limb_t  x0 = xp[0];
      mp_limb_t  y0 = yp[0];
      if (size == 1)
        {
          wp[0] += x0 * y0;
        }
      else if (size == 2)
        {
          mp_limb_t  w1, w0, c;
          umul_ppmm (w1, w0, x0, y0 << GMP_NAIL_BITS);
          ADDC_LIMB (c, wp[0], wp[0], w0 >> GMP_NAIL_BITS);
          wp[1] = w1 + x0*yp[1] + xp[1]*y0 + c;
        }
      else /* size == 3 */
        {
          mp_limb_t  w2, w1, w0, c;
          MUL_333 (w2,w1,w0, xp[2],xp[1],x0, yp[2],yp[1],y0);
          ADDC_LIMB (c, wp[0], wp[0], w0);
          ADDC_LIMB (c, wp[1], wp[1], w0);
          wp[1] = w1 + x0*yp[1] + xp[1]*y0 + c;
        }
    }
  return;
#endif

#if HAVE_NATIVE_mpn_addmul_2
  while (size > 2)
    {
      mpn_addmul_2 (wp, xp, size, yp);
      wp += 2, xp += 2, size -= 2;
    }
  if (size >= 1)
    {
      mp_limb_t  x0 = xp[0];
      mp_limb_t  y0 = yp[0];
      if (size > 1)
        {
          mp_limb_t  c;
          umul_ppmm (w1, w0, x0, y0 << GMP_NAIL_BITS);
          ADDC_LIMB (c, wp[0], wp[0], w0 >> GMP_NAIL_BITS);
          wp[1] = (wp[1] + w1 + x0*yp[1] + xp[1]*y0 + c) & GMP_NUMB_MASK;
        }
      else
        wp[0] = (wp[0] + x0 * y0) & GMP_NUMB_MASK;
    }
  return;
#endif

  while (size > 1)
    {
      mpn_addmul_1 (wp, xp, size, yp[0]);
      wp += 1, yp += 1, size--;
#if HAVE_NATIVE_mpn_addmul_2
      break;
#endif
    }
  wp[0] = (wp[0] + xp[0] * yp[0]) & GMP_NUMB_MASK;
}
